
from pathlib import Path

import aiohttp

from dan.core import aiofiles
from dan.utils import progress


async def fetch_file(url, dest: Path, name: str = None, chunk_size=1024):
    if name is None:
        name = dest.name
    timeout = aiohttp.ClientTimeout(total=30*60, connect=30,
                                    sock_connect=30, sock_read=None)
    async with aiohttp.ClientSession(timeout=timeout) as session:
        async with session.get(url) as resp:
            if resp.status != 200:
                message = await resp.read()
                raise RuntimeError(
                    f'unable to fetch {url}: {message.decode()}')
            size = int(resp.headers.get('content-length', 0))

            with progress(
                desc=f'downloading {name}', total=size // 1024, leave=False, unit='Ko'
            ) as progressbar:
                async with aiofiles.open(dest, mode='wb') as f:
                    async for chunk in resp.content.iter_chunked(chunk_size):
                        await f.write(chunk)
                        progressbar(len(chunk) // 1024)



import collections
import contextlib
from functools import cached_property
import functools
from dan.core.register import MakefileRegister
from dan.core.pathlib import Path
from typing import Any, Callable, Iterable, Union, TypeAlias
import inspect

from dan.core import asyncio, aiofiles, utils, diagnostics as diags
from dan.core.requirements import load_requirements
from dan.core.settings import InstallMode, InstallSettings, safe_load
from dan.core.version import Version
from dan.logging import Logging


class Dependencies:

    def __init__(self, parent: 'Target', public: Iterable = None, private: Iterable = None):
        super().__init__()
        self.parent = parent
        self._public = list()
        self._private = list()
        if public is not None:
            self.update(public, public=True)
        if private is not None:
            self.update(private, public=False)

    @property
    def makefile(self):
        return self.parent.makefile

    def add(self, dependency, public=True):
        content = self._public if public else self._private
        if dependency in content:
            return
        from dan.pkgconfig.package import RequiredPackage
        match dependency:
            case Target() | FileDependency():
                content.append(dependency)
            case type():
                assert issubclass(dependency, Target)
                dep = self.makefile.find(dependency)
                if dep is None:
                    raise RuntimeError(f'cannot find dependency class: {dependency.__name__}')
                content.append(dep)
            case str():
                from dan.pkgconfig.package import Package
                for pkg in Package.all.values():
                    if pkg.name == dependency:
                        content.append(pkg)
                        break
                else:
                    if isinstance(self.parent.source_path, Path) and Path(self.parent.source_path / dependency).exists():
                        content.append(FileDependency(
                            self.parent.source_path / dependency))
                    else:
                        from dan.pkgconfig.package import parse_requirement
                        content.append(parse_requirement(dependency))
            case Path():
                dependency = FileDependency(
                    self.parent.source_path / dependency)
                content.append(dependency)
            case RequiredPackage():
                content.append(dependency)
            case _:
                raise RuntimeError(
                    f'Unhandled dependency {dependency} ({type(dependency)})')

    def update(self, dependencies, public=True):
        match dependencies:
            case Dependencies():
                for dep in dependencies.public:
                    self.add(dep, public=True)
                for dep in dependencies.private:
                    self.add(dep, public=False)
            case collections.abc.Iterable():
                for dep in dependencies:
                    self.add(dep, public=public)
            case _:
                raise RuntimeError('unhandled')

    def __getattr__(self, attr):
        for item in self._public:
            if item.name == attr:
                return item
        for item in self._private:
            if item.name == attr:
                return item
    
    @property
    def public(self):
        return self._public.__iter__()

    @property
    def private(self):
        return self._private.__iter__()

    @property
    def all(self):
        yield from self.private
        yield from self.public

    @property
    def up_to_date(self):
        for item in self.all:
            if not item.up_to_date:
                return False
        return True

    @property
    def modification_time(self):
        t = 0.0
        for item in self.all:
            mt = item.modification_time
            if mt and mt > t:
                t = mt
        return t


TargetDependencyLike: TypeAlias = Union[list['Target'], 'Target']


PathImpl = type(Path())


class FileDependency(PathImpl):
    
    def __init__(self, *args, **kwargs):
        super(PathImpl, self).__init__()

    @property
    def up_to_date(self):
        return self.exists()

    @property
    def modification_time(self):
        return self.stat().st_mtime


class Option:
    def __init__(self, parent: 'Options', fullname: str, default, help: str = None) -> None:
        self.fullname = fullname
        self.name = fullname.split('.')[-1]
        self.__parent = parent
        self.__cache = parent._cache
        self.__default = default
        self.__value = self.__cache.get(self.name, default)
        self.__value_type = type(default)
        self.__help = help if help is not None else 'No description.'

    def reset(self):
        self.value = self.__default

    @property
    def parent(self):
        return self.__parent

    @property
    def cache(self):
        return self.__cache

    @property
    def type(self):
        return self.__value_type

    @property
    def default(self):
        return self.__default

    @property
    def help(self):
        return self.__help

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        value = safe_load(self.fullname, value, self.__value_type)
        if self.__value != value:
            self.__value = value
            self.__cache[self.name] = value


class Options:
    def __init__(self, parent: 'Target', default: dict[str, Any] = dict()) -> None:
        self.__parent = parent
        cache = parent.cache
        if isinstance(parent.cache, dict):
            if not parent.name in cache:
                cache[parent.name] = dict()
            cache = cache[parent.name]
        else:
            cache = parent.cache.data
        if not 'options' in cache:
            cache['options'] = dict()
        self._cache = cache['options']
        self.__items: list[Option] = list()
        self.update(default)

    def add(self, name: str, default_value, help=None):
        if self.get(name, False) is not None:
            raise RuntimeError(f'duplicate options detected ({name})')
        opt = Option(self, f'{self.__parent.fullname}.{name}',
                     default_value, help=help)
        self.__items.append(opt)
        return opt

    def get(self, name: str, parent_lookup = True):
        for o in self.__items:
            if name in {o.name, o.fullname}:
                return o
        if parent_lookup and self.__parent.parent is not None:
            return self.__parent.parent.options.get(name)

    def update(self, options: dict):
        for k, v in options.items():
            help = None
            match v:
                case dict():
                    help = v['help']
                    v = v['default']
                case tuple() | list() | set():
                    help = v[1]
                    v = v[0]
                case _:
                    pass
            if self[k]:
                self[k] = v
            else:
                self.add(k, v, help)
    
    @property
    def sha1(self):
        import hashlib
        sha1 = hashlib.sha1()
        for o in self.__items:
            sha1.update(o.fullname.encode() + str(o.value).encode())
        return sha1.hexdigest()

    def items(self):
        for o in self.__items:
            yield o.name, o.value

    def __getattr__(self, name):
        opt = self.get(name)
        if opt:
            return opt.value

    def __getitem__(self, name):
        opt = self.get(name)
        if opt:
            return opt.value

    def __iter__(self):
        return iter(self.__items)

class Installer:
    def __init__(self, settings: InstallSettings, mode: InstallMode, logger: Logging) -> None:
        self.settings = settings
        self.mode = mode
        self.installed_files = list()
        self._logger = logger
    
    async def _install(self, src: Path|str, dest: Path, subdir: Path = None):
        if subdir is not None:
            dest /= subdir
        if isinstance(src, Path):
            dest /= src.name
            if dest.exists() and dest.younger_than(src):
                self._logger.info('%s is up-to-date', dest)
                self.installed_files.append(dest)
                return
            self._logger.debug('installing: %s', dest)
            await aiofiles.copy(src, dest)
        else:
            dest.parent.mkdir(parents=True, exist_ok=True)
            self._logger.debug('installing: %s', dest)
            async with aiofiles.open(dest, 'w') as f:
                await f.write(src)
        self.installed_files.append(dest)

    @property
    def dev(self):
        return self.mode == InstallMode.dev
    
    async def install_bin(self, src, subdir = None):
        await self._install(src, self.settings.runtime_destination, subdir)

    async def install_shared_library(self, src, subdir = None):
        await self._install(src, self.settings.libraries_destination, subdir)

    async def install_static_library(self, src, subdir = None):
        if not self.dev:
            return
        await self._install(src, self.settings.libraries_destination, subdir)

    async def install_header(self, src, subdir = None):
        if not self.dev:
            return
        await self._install(src, self.settings.includes_destination, subdir)

    async def install_data(self, src, subdir = None, dev=False):
        if dev and not self.dev:
            return
        await self._install(src, self.settings.data_destination, subdir)


class Target(Logging, MakefileRegister, internal=True):
    name: str = None
    fullname: str = None
    description: str = None
    default: bool = True
    installed: bool = False
    output: Path = None
    options: dict[str, Any]|Options = dict()
    provides: Iterable[str] = None

    dependencies: Dependencies = set()
    public_dependencies: set[TargetDependencyLike] = set()
    private_dependencies: set[TargetDependencyLike] = set()

    preload_dependencies: Dependencies = set()

    inherits_version = True
    subdirectory: str = None

    __cache_nop_codec = lambda x: x

    @staticmethod
    def root_cached(name, encode=None, decode=None, get_fn=None):
        """Create new property for cached variable (root scope)"""
        encode = encode or Target.__cache_nop_codec
        decode = decode or Target.__cache_nop_codec
        def get(obj):
            result = obj.makefile.root.cache.data.get(name)
            if result is not None:
                return decode(result)
            elif get_fn is not None:
                result = get_fn(obj)
                if result is not None:
                    obj.makefile.root.cache.data[name] = encode(result)
                return result

        def set(obj, value):
            obj.makefile.root.cache.data[name] = encode(value)
        return property(get, set)

    @staticmethod
    def root_cached_property(name, encode=None, decode=None):
        """Create new property for cached variable (root scope)"""
        def wrapper(get_fn):
            return Target.root_cached(name, encode, decode, get_fn)
        return wrapper
    
    @staticmethod
    def makefile_cached(name, encode=None, decode=None, get_fn=None):
        """Create new property for cached variable (makefile scope)"""
        encode = encode or Target.__cache_nop_codec
        decode = decode or Target.__cache_nop_codec
        def get(obj):
            result = obj.makefile.cache.data.get(name)
            if result is not None:
                return decode(result)
            elif get_fn is not None:
                result = get_fn(obj)
                if result is not None:
                    obj.makefile.cache.data[name] = encode(result)
                return result
        def set(obj, value):
            obj.makefile.cache.data[name] = encode(value)
        return property(get, set)

    @staticmethod
    def makefile_cached_property(name, encode=None, decode=None):
        """Create new property for cached variable (makefile scope)"""
        def wrapper(get_fn):
            return Target.makefile_cached(name, encode, decode, get_fn)
        return wrapper
    
    @staticmethod
    def target_cached(name, encode=None, decode=None, get_fn=None):
        """Create new property for cached variable (target scope)"""
        encode = encode or Target.__cache_nop_codec
        decode = decode or Target.__cache_nop_codec
        def get(obj):
            result = obj.cache.get(name)
            if result is not None:
                return decode(result)
            elif get_fn is not None:
                result = get_fn(obj)
                if result is not None:
                    obj.cache.data[name] = encode(result)
                return result
        def set(obj, value):
            obj.cache[name] = encode(value)
        return property(get, set)
        
    @staticmethod
    def target_cached_property(name, encode=None, decode=None):
        """Create new property for cached variable (target scope)"""
        def wrapper(get_fn):
            return Target.target_cached(name, encode, decode, get_fn)
        return wrapper

    def __init__(self,
                 name: str = None,
                 parent: 'Target' = None,
                 version: str = None,
                 default: bool = None,
                 makefile=None) -> None:
        
        self.parent = parent
        self.__cache: dict = None
        self.__source_path : Path = None

        if name is not None:
            self.name = name

        if self.name is None:
            self.name = self.__class__.__name__
        
        if self.provides is None:
            self.provides = {self.name}
        else:
            self.provides = set(self.provides)

        if default is not None:
            self.default = default

        if parent is not None:
            self.makefile = parent.makefile
            self.fullname = f'{parent.fullname}.{self.name}'

        if makefile:
            self.makefile = makefile

        if self.makefile is None:
            raise RuntimeError('Makefile not resolved')


        if self.fullname is None:
            self.fullname = f'{self.makefile.fullname}.{self.name}'

        self.options = Options(self, self.options)

        if version is not None:
            self._version = version

        if not hasattr(self, '_version'):
            if self.inherits_version:
                self._version = self.makefile.version
            else:
                self._version = None

        if self.description is None:
            self.description = self.makefile.description

        self.other_generated_files: set[Path] = set()

        deps = self.dependencies
        self.dependencies = Dependencies(self, self.public_dependencies, self.private_dependencies)
        self.dependencies.update(deps)
        self.preload_dependencies = Dependencies(
            self, None, self.preload_dependencies)

        self._output: Path = None
        self._build_path = None
        
        if inspect.isclass(self.source_path) and issubclass(self.source_path, Target):
            # delayed resolution
            def _get_source_path(TargetClass, self):
                return self.get_dependency(TargetClass).output
            self.preload_dependencies.add(self.source_path, public=False)
            type(self).source_path = property(functools.partial(_get_source_path, self.source_path))

        if type(self).output != Target.output:
            # hack class-defined output
            #   transform it to classproperty for build_path resolution
            output = self.output
            type(self).output = utils.classproperty(lambda: self.build_path / output)

        self.diagnostics = diags.DiagnosticCollection()

    @property
    def output(self):
        if self._output is None:
            return None
        return self.build_path / self._output

    @property
    def routput(self):
        return self._output
    
    @property
    def version(self):
        version = self._version
        if isinstance(version, Option):
            version = version.value
        if isinstance(version, str):
            version = Version(version)
        return version

    @version.setter
    def version(self, value):
        self._version = value

    @output.setter
    def output(self, path):
        path = Path(path)
        if not path.is_absolute() and self.build_path in path.parents:
            raise RuntimeError(f'output must not be an absolute path within build directory')
        elif path.is_absolute() and self.build_path in path.parents:
            self._output = path.relative_to(self.build_path)
        else:
            self._output = path

    @property
    def is_requirement(self) -> bool:
        return self.makefile.is_requirement

    @property
    def source_path(self) -> Path:
        if self.__source_path is None:
            return self.makefile.source_path
        else:
            return self.__source_path
    
    @source_path.setter
    def source_path(self, value):
        self.__source_path = value

    @property
    def build_path(self) -> Path:
        if self._build_path is not None:
            return self._build_path
        
        build_path = self.makefile.build_path

        if self.subdirectory is not None:
            build_path /= self.subdirectory

        return build_path
        
    
    @property
    def requires(self):
        from dan.pkgconfig.package import RequiredPackage
        return [dep for dep in self.dependencies.all if isinstance(dep, RequiredPackage)]

    @cached_property
    def fullname(self) -> str:
        return f'{self.makefile.fullname}.{self.name}'

    @property
    def cache(self) -> dict:
        if not self.__cache:
            name = self.fullname.removeprefix(self.makefile.fullname + '.')
            if not name in self.makefile.cache.data:
                self.makefile.cache.data[name] = dict()
            self.__cache = self.makefile.cache.data[name]
        return self.__cache
    
    _install_missing_dependencies = True

    def _recursive_dependencies(self, types = None, seen = None):
        if seen is None:
            seen = set()
        for dep in self.dependencies.all:
            if dep in seen:
                continue
            seen.add(dep)
            if types is None or isinstance(dep, types):
                yield dep
                if isinstance(dep, Target):
                    yield from dep._recursive_dependencies(types, seen)

    @property
    @contextlib.contextmanager
    def skip_missing_dependencies(self):
        self._install_missing_dependencies = False
        yield
        self._install_missing_dependencies = True

    async def __load_unresolved_dependencies(self, install=None):
        if install is None:
            install = self._install_missing_dependencies
        if len(self.requires) > 0:
            self.dependencies.update(await load_requirements(self.requires, makefile=self.makefile, logger=self, install=install))

    @asyncio.cached
    async def preload(self):
        self.trace('preloading...')

        async with asyncio.TaskGroup(f'building {self.name}\'s preload dependencies') as group:
            group.create_task(self.__load_unresolved_dependencies())
            for dep in self.preload_dependencies.all:
                group.create_task(dep.build())

        async with asyncio.TaskGroup(f'preloading {self.name}\'s target dependencies') as group:
            for dep in self.target_dependencies:
                group.create_task(dep.preload())

        return await asyncio.may_await(self.__preload__())

    @asyncio.cached
    async def load_dependencies(self):
        async with asyncio.TaskGroup(f'loading {self.name}\'s dependencies') as group:
            group.create_task(self.__load_unresolved_dependencies(install=False))

    @asyncio.cached
    async def initialize(self):
        await self.preload()
        self.trace('initializing...')

        if isinstance(self.version, Option):
            self.version = self.version.value

        async with asyncio.TaskGroup(f'initializing {self.name}\'s target dependencies') as group:
            for dep in self.target_dependencies:
                group.create_task(dep.initialize())

        return await asyncio.may_await(self.__initialize__())

    @property
    def modification_time(self):
        output = self.build_path / f'{self.name}.stamp' if self.output is None else self.output  
        return output.stat().st_mtime if output.exists() else 0.0

    @cached_property
    def up_to_date(self):
        output = self.build_path / f'{self.name}.stamp' if self.output is None else self.output
        if output and not output.exists():
            return False
        elif not self.dependencies.up_to_date:
            return False
        elif self.dependencies.modification_time > self.modification_time:
            return False
        elif 'options_sha1' in self.cache and self.cache['options_sha1'] != self.options.sha1:
            return False
        return True

    async def _build_dependencies(self):
        async with asyncio.TaskGroup(f'building {self.name}\'s target dependencies') as group:
            for dep in self.target_dependencies:
                group.create_task(dep.build())

    @asyncio.cached
    async def build(self):
        await self.initialize()

        await self._build_dependencies()

        result = await asyncio.may_await(self.__prebuild__())

        if self.up_to_date:
            self.trace('up to date !')
            return
        elif self.output is not None and self.output.exists():
            self.debug('outdated !')

        with utils.chdir(self.build_path):
            self.debug('building...')
            if diags.enabled:
                self.diagnostics.clear()
                
            result = await asyncio.may_await(self.__build__())
            if self.output is None:
                (self.build_path / f'{self.name}.stamp').touch()
            self.cache['options_sha1'] = self.options.sha1
            self.trace('built')
            return result

    @property
    def target_dependencies(self):
        return [t for t in {*self.dependencies.all, *self.preload_dependencies.all} if isinstance(t, Target)]

    @property
    def file_dependencies(self):
        return [t for t in self.dependencies.all if isinstance(t, FileDependency)]

    @asyncio.cached
    async def clean(self):
        await self.initialize()
        async with asyncio.TaskGroup(f'cleaning {self.name} outputs') as group:
            output = self.build_path / f'{self.name}.stamp' if self.output is None else self.output
            if output and output.exists():
                self.info('debug...')
                if output.is_dir():
                    group.create_task(aiofiles.rmtree(output, force=True))
                else:
                    group.create_task(aiofiles.os.remove(output))
            for f in self.other_generated_files:
                if f.exists():
                    group.create_task(aiofiles.os.remove(f))
            group.create_task(asyncio.may_await(self.__clean__()))

    @asyncio.cached(unique = True)
    async def install(self, settings: InstallSettings, mode: InstallMode):
        await self.build()

        self.debug('installing %s to %s', self.name, settings.destination)

        installer = Installer(settings, mode, self)
        await self.__install__(installer)
        return installer.installed_files


    def __get_dependency(self, dep: str | type, recursive=True) -> TargetDependencyLike:
        """Search for dependency"""
        if isinstance(dep, str):
            def check(d): return d.name == dep
        else:
            def check(d): return isinstance(d, dep)
        for dependency in self.dependencies.all:
            if check(dependency):
                return dependency
        for dependency in self.preload_dependencies.all:
            if check(dependency):
                return dependency
        if recursive:
            # not found... look for dependencies' dependencies
            for target in self.target_dependencies:
                dependency = target.__get_dependency(dep)
                if dependency is not None:
                    return dependency

    def get_dependency(self, dep: str | type, recursive=True) -> TargetDependencyLike:
        dependency = self.__get_dependency(dep, recursive)
        from dan.core.requirements import RequiredPackage
        match dependency:
            case RequiredPackage():
                if dependency.target is not None:
                    return dependency.target
                else:
                    return dependency
            case _:
                return dependency

    async def __preload__(self):
        ...

    async def __initialize__(self):
        ...

    async def __prebuild__(self):
        ...

    async def __build__(self):
        ...

    async def __install__(self, installer: Installer):
        if installer.dev:
            if len(self.utils) > 0:
                body = str()
                for fn in self.utils:
                    tmp = inspect.getsourcelines(fn)[0]
                    tmp[0] = f'\n\n@self.utility\n'
                    body += '\n'.join(tmp)
                await installer.install_data(body, f'dan/{self.name}.py')

    async def __clean__(self):
        ...

    @utils.classproperty
    def utils(cls) -> list:
        utils_name = f'_{cls.__name__}_utils__'
        if not hasattr(cls, utils_name):
            setattr(cls, utils_name, list())
        return getattr(cls, utils_name)

    @classmethod
    def utility(cls, fn: Callable):
        cls.utils.append(fn)
        name = fn.__name__
        makefile = cls.get_static_makefile()
        if makefile is not None:
            inst = makefile.find(cls)
            fn = functools.partial(fn, inst)
        setattr(cls, name, fn)
        return fn

    async def run(self, command, cwd=None, env=None, **kwargs):
        from dan.core.runners import async_run
        kwargs['logger'] = self
        if cwd is None:
            cwd = self.build_path
        if env is None:
            env = getattr(self, 'env', None)
        return await async_run(command, cwd=cwd, env=env, **kwargs)



import dataclasses
from enum import Enum
import types as typs
from click import *

import inspect
import asyncio

import dan.core.typing as t
from dan import logging

class AsyncContext(Context):
    def invoke(__self, __callback, *args, **kwargs):
        ret = super().invoke(__callback, *args, **kwargs)
        if inspect.isawaitable(ret):
            loop = asyncio.get_event_loop()
            if loop.is_running():
                return ret  # must be awaited
            return loop.run_until_complete(ret)
        else:
            return ret


BaseCommand.context_class = AsyncContext

logger = logging.getLogger('cli')


class SettingsParamType(ParamType):
    def __init__(self, setting_cls) -> None:
        self.setting_cls = setting_cls
        self.fields = dataclasses.fields(self.setting_cls)
        super().__init__()

    def shell_complete(self, ctx, param, incomplete):
        from click.shell_completion import CompletionItem
        completions: list[str] = list()
        def gen_comps(fields, parts: list[str], prefix=''):
            part = parts[0]
            if part.startswith('"'):
                part = part[1:]
            value = None
            if '=' in part:
                part, value = part.split('=')
            elif '+=' in part or '-=' in part:
                part, value = part.split(part[:-2])
            for field in fields:
                if part.startswith(field.name):
                    type = field.type
                    if isinstance(type, typs.GenericAlias):
                        type = t.get_origin(type)
                    elif t.is_optional(type):
                        type = t.get_args(type)[0]
                    if dataclasses.is_dataclass(type):
                        subfields = dataclasses.fields(type)
                        subparts = parts[1:]
                        gen_comps(subfields, subparts, f'{field.name}.')
                    elif issubclass(type, (set, list)):
                        if part.endswith(('+', '-')):
                            completions.append(CompletionItem(f'{prefix}{field.name}{part[-1]}=', type='nospace'))
                        else:
                            for s in ('=', '+=', '-='):
                                completions.append(CompletionItem(f'{prefix}{field.name}{s}', type='nospace'))
                    elif issubclass(type, Enum):
                        for evalue in type._member_names_:
                            if value is None or evalue.startswith(value):
                                completions.append(CompletionItem(f'"{prefix}{field.name}={evalue}"'))
                    else:
                        completions.append(CompletionItem(f'{prefix}{field.name}=', type='nospace'))
                    break
                elif field.name.startswith(part):
                    type = field.type
                    if isinstance(type, typs.GenericAlias):
                        type = t.get_origin(type)
                    if dataclasses.is_dataclass(type):
                        completions.append(CompletionItem(f'{prefix}{field.name}.', type='nospace'))
                    else:
                        completions.append(CompletionItem(f'{prefix}{field.name}', type='nospace'))
        gen_comps(fields=self.fields, parts=incomplete.split('.'))
        return completions

class OptionsParamType(ParamType):
    def shell_complete(self, ctx: AsyncContext, param, incomplete):
        from click.shell_completion import CompletionItem
        from dan.make import Make
        from dan.core.asyncio import sync_wait
        build_path = ctx.params['build_path']
        make = Make(build_path, quiet=True)
        sync_wait(make.initialize())
        
        comps = []
        for opt in make.all_options:
            if opt.fullname.startswith(incomplete):
                comps.append(CompletionItem(opt.fullname, type='nospace'))
        
        return comps


class TargetParamType(ParamType):
    def __init__(self, target_types=None) -> None:
        from dan.core.target import Target
        if target_types is None:
            self.target_types = (Target)
        else:
            self.target_types = tuple(target_types)
        super().__init__()

    def shell_complete(self, ctx: AsyncContext, param, incomplete):
        from click.shell_completion import CompletionItem
        from dan.make import Make
        from dan.core.asyncio import sync_wait
        build_path = ctx.params['build_path']
        make = Make(build_path, quiet=True)
        sync_wait(make.initialize())

        
        comps = []
        for target in make.root.all_targets:
            if isinstance(target, self.target_types) and target.fullname.startswith(incomplete):
                comps.append(CompletionItem(target.fullname, type='nospace'))
        
        return comps


class ToolchainParamType(ParamType):
    def shell_complete(self, ctx: AsyncContext, param, incomplete):
        from dan.cxx.detect import get_toolchains
        from click.shell_completion import CompletionItem
        toolchains = get_toolchains(create=False)["toolchains"]
        
        comps = []
        for name in toolchains.keys():
            if name.startswith(incomplete):
                comps.append(CompletionItem(name))

        return comps



from dataclasses import dataclass, field
from pathlib import Path
from dataclasses_json import DataClassJsonMixin, config, LetterCase
from enum import Enum
import typing as t


def hidden_field(*args, **kwargs):
    """Return an object to identify dataclass fields, but is will not be serialized.
    
    See: dataclasses.field"""
    metadata = config(kwargs.pop('metadata', None), exclude=lambda _:True)
    return field(*args, metadata=metadata, **kwargs)


enabled = False

class Severity(Enum):
    ERROR = 0
    WARNING = 1
    INFO = 2
    HINT = 3

@dataclass(init=False)
class Uri:
    """Javascript-friendly URI"""
    scheme: str = 'file'
    path: str
    fragment: str = ''

    def __init__(self, path: str|Path) -> None:
        self.path = str(path)

@dataclass
class Position:
    line: int = 0
    character: int = 0

@dataclass
class Range:
    start: Position = field(default_factory=Position)
    end: Position = field(default_factory=Position)


@dataclass
class Location:
    uri: Uri
    range: Range


@dataclass
class RelatedInformation:
    location: Location
    message: str


@dataclass
class Diagnostic(DataClassJsonMixin):
    dataclass_json_config = config(letter_case=LetterCase.CAMEL)['dataclasses_json']
    message: str
    range: Range = field(default_factory=Range)
    severity: Severity = Severity.ERROR
    code: t.Optional[str|int] = None
    source: t.Optional[str] = None
    related_information: t.Optional[list[RelatedInformation]] = None
    filename: t.Optional[str] = hidden_field(default=None)


class DiagnosticCollection(dict[str, list[Diagnostic]], DataClassJsonMixin):
    
    def __setitem__(self, key: str, value: list[Diagnostic]) -> None:
        match value:
            case list():
                return super().__setitem__(key, value)
            case Diagnostic():
                if not key in self:
                    super().__setitem__(key, list())
                return self[key].append(value)
            case _:
                raise ValueError(f'Unallowed assignment: {type(value)}')

    def insert(self, diagnostics: list[Diagnostic],  default_key: str):
        for diagnostic in diagnostics:
            self[default_key if diagnostic.filename is None else diagnostic.filename] = diagnostic



from pathlib import Path
from dan.core.asyncio import sync_wait
from dan.core.makefile import MakeFile

from dan.cxx.detect import get_dan_path
from dan.core.target import Target
from dan.core.runners import CommandError, async_run
from dan.core import aiofiles
from dan.core.cache import Cache

from dataclasses_json import dataclass_json
from dataclasses import dataclass, field

import typing as t


@dataclass
class RepositoryConfig:
    name: str
    url: str
    branch: str = 'main'


@dataclass
class GitHubConfig:
    api_token: t.Optional[str] = None


@dataclass_json
@dataclass
class RepositoriesSettings:
    github: GitHubConfig = field(default_factory=GitHubConfig)
    repositories: list[RepositoryConfig] = field(default_factory=lambda: [
        RepositoryConfig('dan.io', 'https://github.com/Garcia6l20/dan.io.git'),
    ])

    def get(self, name):
        for config in self.repositories:
            if config.name == name:
                return config

    @property
    def default(self):
        return self.repositories[0]


class RepositoriesConfig(Cache[RepositoriesSettings]):
    indent = 2


_repo_settings: RepositoriesConfig = None
_repo_instances: dict[str, 'PackageRepository'] = dict()


def _get_settings() -> RepositoriesSettings:
    global _repo_settings
    if _repo_settings is None:
        _repo_settings = RepositoriesConfig(
            get_dan_path() / 'repositories.json')
        if not _repo_settings.path.exists():
            sync_wait(_repo_settings.save(force=True))
    return _repo_settings.data


class PackageRepository(Target, internal=True):

    # never up-to-date
    up_to_date = False

    def __init__(self, name: RepositoryConfig, *args, **kwargs):
        self.repo_data = _get_settings().get(name)
        super().__init__(name, *args, **kwargs)
        self.output = get_dan_path() / 'repositories' / self.name
        self.toolchain = self.context.get('cxx_target_toolchain')
        self.pkgs_root = get_packages_path() / self.toolchain.system / \
            self.toolchain.arch / self.toolchain.build_type.name
        self._package_makefile = None

    async def __build__(self):
        if not self.output.exists():
            try:
                self.output.parent.mkdir(exist_ok=True, parents=True)
                await async_run(f'git clone -b {self.repo_data.branch} {self.repo_data.url} {self.name}', logger=self, cwd=self.output.parent)

            except Exception as e:
                await aiofiles.rmtree(self.output)
                raise e
        else:
            try:
                await async_run(f'git pull -q', logger=self, cwd=self.output)
            except CommandError:
                self.warning('cannot update %s', self.name)

    @property
    def pkgs_makefile(self) -> MakeFile:
        if self._package_makefile is None:
            from dan.core.include import load_makefile
            root = self.output / 'packages'
            with self.makefile.context:
                self._package_makefile = load_makefile(root / 'dan-build.py',
                                                       f'{self.name}.packages',
                                                       requirements=None,
                                                       build_path=self.build_path / self.name,
                                                       parent=self.makefile,
                                                       is_requirement=True)
                self._package_makefile.pkgs_path = self.pkgs_root

        return self._package_makefile

    @property
    def installed(self) -> dict[str, Target]:
        pkgs = self.pkgs_makefile
        return {f'{lib.makefile.name}:{lib.name}@{self.name}': lib for lib in pkgs.all_installed}

    def find(self, name: str, package: str) -> tuple[MakeFile, Target]:
        for lib in self.pkgs_makefile.all_installed:
            if name in lib.provides and (package is None or lib.makefile.name == package):
                return lib.makefile, lib
        return None, None


def get_repo_instance(repo_name: str, makefile=None) -> PackageRepository:
    if makefile is None:
        from dan.core.include import context
        makefile = context.root

    if repo_name is None:
        repo_name = _get_settings().default.name

    if not repo_name in _repo_instances:
        _repo_instances[repo_name] = PackageRepository(
            repo_name, makefile=makefile)
    return _repo_instances[repo_name]


def get_all_repo_instances(makefile=None) -> list[PackageRepository]:
    for repo in _get_settings().repositories:
        get_repo_instance(repo.name, makefile)
    return _repo_instances.values()


def get_packages_path() -> Path:
    return get_dan_path() / 'packages'



from dan.core.pathlib import Path
from dan.core.settings import InstallMode
from tests import PyMakeBaseTest


class CXXSimpleTest(PyMakeBaseTest):
    def __init__(self, methodName: str = None) -> None:
        super().__init__('cxx/simple', methodName)

    async def test_build(self):

        ########################################
        async with self.section("base build", clean=True) as make:
            target = make.root.find('simple')
            await target.initialize()
            self.assertFalse(target.output.exists())
            await target.build()
            self.assertTrue(target.output.exists())
            modified_at = target.output.modification_time


        ########################################
        async with self.section("no-modification => no-rebuild") as make:
            target = make.root.find('simple')
            await target.build()
            self.assertTrue(target.output.exists())
            self.assertEqual(modified_at, target.output.modification_time,
                            "no modifications should NOT trigger a re-build")

            # update source
            src: Path = target.source_path / list(target.sources)[0]
            src.utime()

        ########################################
        async with self.section("source modification => rebuild") as make:
            target = make.root.find('simple')
            await target.build()
            self.assertTrue(target.output.younger_than(modified_at),
                            "a source modification should trigger a re-build")


    async def test_option_change(self):

        ########################################
        async with self.section("base build", clean=True) as make:
            target = make.root.find('simple')
            await target.initialize()
            self.assertFalse(target.output.exists())
            await target.build()
            self.assertTrue(target.output.exists())
            modified_at = target.output.modification_time
            sha1 = target.options.sha1
            
            # change option
            greater = target.options.get('greater')
            expected_output = '=== test ==='
            greater.value = expected_output
            self.assertNotEqual(sha1, target.options.sha1)

        ########################################
        async with self.section("option modification => rebuild") as make:
            target = make.root.find('simple')
            await target.build()
            self.assertTrue(target.output.younger_than(modified_at),
                            "an option change should trigger a re-build")
            modified_at = target.output.modification_time
            out, err, rc = await target.execute()
            self.assertEqual(rc, 0)
            self.assertEqual(out.strip(), f'{expected_output} !')

    async def test_install(self):

        ########################################
        async with self.section("install",
                                settings=[f"install.destination={self.build_path}/dist"]
                                ) as make:
            await make.initialize()
            await make.install(InstallMode.user)
            bins = list((self.build_path / 'dist/bin').glob('*'))
            self.assertEqual(len(bins), 1)
            self.assertEqual(bins[0].stem, 'simple')



from dan.core.version import Version
from dan.core.find import (
    find_executable,
    find_executables,
    find_file,
    find_files,
    find_include_path,
    find_library,
    )
from dan.core.runners import async_run, sync_run
from dan.core.pm import re_fullmatch, re_match, re_search
from dan.core.progress import progress
from dan.cli import user as user_cli



from pathlib import Path
from tests import PyMakeBaseTest

from dan.cxx.toolchain import CompilationFailure, LinkageFailure
from dan.core.asyncio import ExceptionGroup


base_path = Path(__file__).parent / 'errors' / 'cxx'

class CXXSimpleErrors(PyMakeBaseTest):
    def __init__(self, methodName: str = None) -> None:
        super().__init__(methodName=methodName, source_path=base_path)

    async def test_invalid_syntax(self):
        async with self.section('invalid-syntax', targets=['InvalidSyntax'], clean=True, diags=True) as make:
            try:
                await make.build()
                self.fail('No exception raised')
            except ExceptionGroup as err:
                # Note: CompilationFailure always within a group, because all objects are built in parallel
                self.assertEqual(len(err.errors), 1, 'Error not detected')
                err : CompilationFailure = err.errors.pop()
                self.assertTrue(isinstance(err, CompilationFailure))
                self.assertEqual(err.sourcefile.name, 'invalid-syntax.cpp')
                diags = list(err.diags)
                self.assertGreaterEqual(len(diags), 1, 'Error not detected')
            except RuntimeError as err:
                self.fail(f'Wrong exception raised: {err} ({type(err)})')

    async def test_no_main(self):
        async with self.section('no-main', targets=['NoMain'], clean=True, diags=True) as make:
            try:
                await make.build()
                self.fail('No exception raised')
            except LinkageFailure as err:
                diags = list(err.diags)
                self.assertEqual(len(diags), 1, 'Error not detected')
                diag = diags[0]
                if make.toolchain.type == 'msvc':
                    self.assertTrue(diag.code == 'LNK1561')
                else:
                    self.assertTrue('undefined reference' in diag.message)
                    self.assertTrue('main' in diag.message)
            except RuntimeError:
                self.fail('Wrong exception raised')

    async def test_undefined_reference(self):
        async with self.section('undefined-reference', targets=['UndefinedReference'], clean=True, diags=True) as make:
            try:
                await make.build()
                self.fail('No exception raised')
            except LinkageFailure as err:
                diags = err.diags
                self.assertEqual(len(diags), 2, 'Error not detected')
                undefined_vars = ['undefined1', 'undefined2']
                for diag in diags:
                    # self.assertTrue(Path(error.filename).name == 'undefined-reference.cpp')
                    if make.toolchain.type == 'msvc':
                        self.assertTrue(diag.code == 'LNK2001')
                    else:
                        # self.assertTrue(diag.function == 'main')
                        self.assertTrue('undefined reference' in diag.message)
                    # Note: may not be ordered
                    self.assertTrue(any(var_name in diag.message for var_name in undefined_vars))
            except RuntimeError:
                self.fail('Wrong exception raised')



from pathlib import Path
from dan.core.settings import InstallMode, InstallSettings
from dan.core.target import Target, FileDependency, Installer
from dan.core.runners import async_run
from dan.core import aiofiles, asyncio
from dan.core.find import find_file
from dan.cxx import Toolchain

import typing as t
import os

import platform

async def get_ninja():
    from dan.cxx.detect import get_dan_path
    match platform.system():
        case 'Windows':
            suffix = '.exe'
        case _:
            suffix = ''
    bin_path = get_dan_path() / 'os-utils' / 'bin'
    ninja_path = bin_path / f'ninja{suffix}'
    ninja_version = '1.11.1'
    if not ninja_path.exists():
        match platform.system():
            case 'Windows':
                name = 'ninja-win'
            case 'Linux':
                name = 'ninja-linux'
            case 'Darwin':
                name = 'ninja-mac'
        import tempfile
        import zipfile
        import stat
        from dan.utils.net import fetch_file
        with tempfile.TemporaryDirectory(prefix=f'dan-ninja-') as tmp_dest:
            tmp_dest = Path(tmp_dest)
            archive_name = f'{name}.zip'
            await fetch_file(f'https://github.com/ninja-build/ninja/releases/download/v{ninja_version}/{archive_name}', tmp_dest / archive_name)
            with zipfile.ZipFile(tmp_dest / archive_name) as f:
                f.extractall(bin_path)
            ninja_path.chmod(ninja_path.stat().st_mode | stat.S_IEXEC)
    return ninja_path



class Project(Target, internal=True):

    cmake_targets: list[str] = None
    cmake_config_definitions: dict[str, str] = dict()
    cmake_patch_debug_postfix: list = None
    cmake_options: dict[str, tuple[str, t.Any, str]] = None
    cmake_generator: str = 'Ninja'
    cmake_subdirectory: str = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cmake_cache_dep = FileDependency(self.build_path / 'CMakeCache.txt')
        self.dependencies.add(self.cmake_cache_dep)
        self.toolchain : Toolchain = self.context.get('cxx_target_toolchain')
        self.__env = None

    def get_env(self):
        if self.__env is None:
            env = self.toolchain.env
            from dan.cxx.detect import get_dan_path
            epath = env.get('PATH', os.environ['PATH']).split(os.pathsep)
            epath.insert(0, str(get_dan_path() / 'os-utils' / 'bin'))
            env['PATH'] = os.pathsep.join(epath)
            self.__env = env
        return self.__env


    async def _cmake(self, *cmake_args, **kwargs):
        return await async_run(['cmake', *cmake_args], logger=self, cwd=self.build_path, **kwargs, env=self.get_env())

    @property
    def _target_args(self):
        targets_args = []
        if self.cmake_targets is not None:
            for target in self.cmake_targets:
                targets_args.extend(('-t', target))
        return targets_args

    async def __initialize__(self):
        if self.cmake_options is not None:
            for name, (cmake_name, default, help) in self.cmake_options.items():
                opt = self.options.add(name, default, help)
                setattr(opt, 'cmake_name', cmake_name)
        return await super().__initialize__()

    async def __build__(self):
        cmake_options = dict()
        for opt in self.options:
            if hasattr(opt, 'cmake_name'):
                value = opt.value
                if isinstance(value, bool):
                    value = 'ON' if value else 'OFF'
                cmake_options[opt.cmake_name] = value

        cmake_options['CMAKE_PREFIX_PATH'] = self.makefile.root.pkgs_path.as_posix()

        base_opts = []
        if self.cmake_generator.startswith('Ninja'):
            ninja = await get_ninja()
            base_opts.extend((f'-G{self.cmake_generator}', f'-DCMAKE_MAKE_PROGRAM={ninja.as_posix()}'))
        else:
            raise RuntimeError('Only Ninja generators are currently supported')
        
        if 'multi' in self.cmake_generator.lower():
            base_opts.append(f'-DCMAKE_CONFIGURATION_TYPES={self.toolchain.build_type.name.title()}')
        else:
            base_opts.append(f'-DCMAKE_BUILD_TYPE={self.toolchain.build_type.name.title()}')

        source_path = self.source_path
        if self.cmake_subdirectory:
            source_path /= self.cmake_subdirectory

        await self._cmake(
            source_path,
            *base_opts,
            f'-DCMAKE_C_COMPILER={self.toolchain.cc.as_posix()}',
            f'-DCMAKE_CXX_COMPILER={self.toolchain.cxx.as_posix()}',
            *[f'-D{k}={v}' for k, v in self.cmake_config_definitions.items()],
            *[f'-D{k}={v}' for k, v in cmake_options.items()]
        )
        await self._cmake('--build', '.', '--parallel', *self._target_args)
    
    @asyncio.cached(unique = True)
    async def install(self, settings: InstallSettings, mode: InstallMode):
        self.cmake_config_definitions['CMAKE_INSTALL_PREFIX'] = settings.destination
        return await super().install(settings, mode)

    async def __install__(self, installer: Installer):
        await self._cmake('--install', '.')

        await super().__install__(installer)
        async with aiofiles.open(self.build_path / 'install_manifest.txt') as manifest_file:
            manifest = await manifest_file.readlines()
            
        if self.cmake_patch_debug_postfix is not None and self.toolchain.build_type.is_debug_mode:
            # fix: no 'd' postfix in MSVC pkgconfig
            seach_paths = [
                installer.settings.data_destination / 'pkgconfig',
                installer.settings.libraries_destination / 'pkgconfig',
            ]
            for provided in self.provides:
                pc_file = find_file(rf'{provided}\.pc', paths=seach_paths)
                self.debug('patching %s', pc_file)
                for name in self.cmake_patch_debug_postfix:
                    await aiofiles.sub(pc_file, rf'-l{name}(\s)', rf'-l{name}d\g<1>')

        installer.installed_files.extend(manifest)



import json
import os
from dan.core import asyncio

from dan.core.pm import re_match
from dan.cxx.toolchain import Toolchain

from dan.make import Make
from dan.cxx.targets import CXXObject
from dan.logging import Logging
from dan.core.utils import unique

def get_intellisense_mode(toolchain : Toolchain):
    mode = list()
    if toolchain.system is not None:
        if toolchain.system.startswith('msys'):
            mode.append('windows')
        else:
            mode.append(toolchain.system)
    mode.append(toolchain.type)
    mode.append(toolchain.arch)
    return '-'.join(mode)

class Code(Logging):
    fullname = 'code'
    def __init__(self, make: Make) -> None:
        self.make = make

    def get_test_suites(self, pretty):
        from dan.core.include import MakeFile
        from dan.core.test import Test, Case
        from dan.cxx import Executable

        def make_inner_test_info(test: Test, case: Case):
            basename = test.basename(case)
            out, err = test.outs(case)
            ident = f'{test.fullname}:{case.name}' if case.name is not None else test.fullname
            info = {
                'type': 'test',
                'id': ident,
                'label': basename,
                'debuggable': False,
                'target': test.executable.fullname,
                'out': str(out),
                'err': str(err),
            }
            if isinstance(test.executable, Executable):
                info['debuggable'] = True
                if case.file:
                    info['file'] = str(case.file)
                elif test.file:
                    info['file'] = str(test.file)
                else:
                    info['file'] = str(
                        test.executable.source_path / test.executable.sources[0])

                if case.lineno:
                    info['line'] = case.lineno
                elif test.lineno:
                    info['line'] = test.lineno

                if test.workingDir:
                    info['workingDirectory'] = str(test.workingDir)
                else:
                    info['workingDirectory'] = str(test.executable.build_path)

                if len(case.args) > 0:
                    info['args'] = [str(a) for a in case.args]

            return info

        def make_test_info(test: Test):
            if len(test) == 0:
                raise RuntimeError(f'Test: {test.name} has not test')
            if len(test) == 1:
                return make_inner_test_info(test, test.cases[0])
            else:
                return {
                    'type': 'suite',
                    'id': test.fullname,
                    'label': test.name,
                    'children': [make_inner_test_info(test, case) for case in test.cases]
                }



        def make_suite_info(mf: MakeFile):
            if len(mf.tests) == 0 and mf.children == 0:
                return None
            children = list()
            for test in mf.tests:
                children.append(make_test_info(test))

            for child in mf.children:
                child_suite = make_suite_info(child)
                if child_suite is not None:
                    children.append(child_suite)

            if len(children) > 0:
                return {
                    'type': 'suite',
                    'id': mf.fullname,
                    'label': mf.name,
                    'children': children
                }

        return json.dumps(make_suite_info(self.make.context.root), indent=2 if pretty else None)

    async def _init_target(self, target):
        with target.skip_missing_dependencies:
            await target.initialize()

    async def _make_source_configuration(self, target: CXXObject):
            # interface:
            #   - includePath: string[]
            #   - defines: string[]
            #   - intelliSenseMode?: "linux-clang-x86" | "linux-clang-x64" | "linux-clang-arm" | "linux-clang-arm64" | "linux-gcc-x86" | "linux-gcc-x64" | "linux-gcc-arm" | "linux-gcc-arm64" | "macos-clang-x86" | "macos-clang-x64" | "macos-clang-arm" | "macos-clang-arm64" | "macos-gcc-x86" | "macos-gcc-x64" | "macos-gcc-arm" | "macos-gcc-arm64" | "windows-clang-x86" | "windows-clang-x64" | "windows-clang-arm" | "windows-clang-arm64" | "windows-gcc-x86" | "windows-gcc-x64" | "windows-gcc-arm" | "windows-gcc-arm64" | "windows-msvc-x86" | "windows-msvc-x64" | "windows-msvc-arm" | "windows-msvc-arm64" | "msvc-x86" | "msvc-x64" | "msvc-arm" | "msvc-arm64" | "gcc-x86" | "gcc-x64" | "gcc-arm" | "gcc-arm64" | "clang-x86" | "clang-x64" | "clang-arm" | "clang-arm64";
            #   - standard?: "c89" | "c99" | "c11" | "c17" | "c++98" | "c++03" | "c++11" | "c++14" | "c++17" | "c++20" | "gnu89" | "gnu99" | "gnu11" | "gnu17" | "gnu++98" | "gnu++03" | "gnu++11" | "gnu++14" | "gnu++17" | "gnu++20";
            #   - forcedInclude?: string[];
            #   - compilerPath?: string;
            #   - compilerArgs?: string[];
            #   - windowsSdkVersion?: string;
            includes = await target.toolchain.get_default_include_paths()
            defines = [f'{k}={v}' for k, v in (await target.toolchain.get_default_defines()).items()]            
            await self._init_target(target)
            for flag in target.private_cxx_flags:
                match re_match(flag):
                    case r'[/-]I:?(.+)' as m:
                        includes.append(m[1])
                    case r'[/-]D:?(.+)' as m:
                        defines.append(m[1])

            config = {
                'includePath': [os.path.normcase(i) for i in unique(includes)],
                'defines': defines,
                'compilerPath': os.path.normcase(target.toolchain.cxx),
                'intelliSenseMode': get_intellisense_mode(target.toolchain),
                'compilerArgs': target.cxx_flags,
            }
            if target.cpp_std is not None:
                config['standard'] = f'c++{target.cpp_std}'
            return config

    async def get_sources_configuration(self, sources):
        result = list()
        for source in sources:
            target = await self.make.target_of(source)
            if target:
                result.append({
                    'uri': source,
                    'configuration': await self._make_source_configuration(target)
                })
        return json.dumps(result)
    
    async def get_workspace_browse_configuration(self):
        # interface:
        #   - browsePath: string[];
        #   - compilerPath?: string;
        #   - compilerArgs?: string[];
        #   - standard?: see above
        #   - windowsSdkVersion?: string
        from dan.cxx import target_toolchain as toolchain
        from dan.cxx.targets import CXXTarget
        cpp_std = 11
        browse_path = set()
        compiler_args = set()
        cxx_targets = [t for t in self.make.root.all_targets if isinstance(t, CXXTarget)]
        async with asyncio.TaskGroup("initializing cxx targets") as g:
            for target in cxx_targets:
                g.create_task(self._init_target(target))

        for target in cxx_targets:
            browse_path.update(target.includes.private_raw)
            browse_path.update(target.includes.public_raw)
            compiler_args.update(target.cxx_flags)
            if target.cpp_std is not None and target.cpp_std > cpp_std:
                cpp_std = target.cpp_std

        from dan.pkgconfig.package import get_packages_cache
        for package in get_packages_cache().values():
            compiler_args.update(package.cxx_flags)


        result = {
            'browsePath': [os.path.normcase(p) for p in browse_path],
            'compilerPath': str(toolchain.cxx),
            'compilerArgs': list(compiler_args),
            'standard': f'c++{cpp_std}'
        }
        return json.dumps(result)



from pathlib import Path
from dan.cxx import Executable, Library
from dan.core import asyncio
from dan import logging

import typing as t

import os
import functools
import glob
from elftools.elf import elffile


log = logging.getLogger(__name__)

def find_file_ignore_case(dirname, filename):
    if not os.path.exists(dirname):
        return None
    (root, _, filenames) = next(os.walk(dirname))
    filename_lower = filename.lower()
    for f in filenames:
        if f.lower() == filename_lower:
            return os.path.join(root, f)
    return None


if os.name == 'nt':
    import pefile

    def get_pe_deps(pe_data):
        deps = []
        if hasattr(pe_data, 'DIRECTORY_ENTRY_IMPORT'):
            for imp in pe_data.DIRECTORY_ENTRY_IMPORT:
                deps.append(imp.dll.decode())
        return deps


    def find_dll(dll, dll_lookup_dirs, arch):
        for dir in dll_lookup_dirs:
            dll_path = find_file_ignore_case(dir, dll)
            if not dll_path:
                continue
            pe_data = pefile.PE(dll_path)
            if pe_data.FILE_HEADER.Machine == arch:
                return (dll, dll_path, get_pe_deps(pe_data))
        return (dll, None, None)

    async def pe_tree(pe, lookup_dirs):
        libs = {} # stores all dlls we encounter, specifically {(basename(dll)).lower(): abspath(dll)}
        deps = {} # stores pe -> [dll] relations, specifically {abspath(pe): [basename(pe's direct dll dep)]}

        data = pefile.PE(pe)
        arch = data.FILE_HEADER.Machine

        async def _dep_tree(bin, bin_deps):
            bin = os.path.abspath(bin)
            if bin in deps:
                # aleady processed before
                return
            deps[bin] = bin_deps
            results = []
            # pefile.PE() takes very long time to run, so use multiprocessing to speed it up        
            async with asyncio.TaskGroup() as group:
                for dll in bin_deps:
                    if dll.lower() not in libs:
                        group.create_task(asyncio.async_wait(find_dll, dll, lookup_dirs, arch))

            results = group.results()
            for dll, dll_path, _deps in results:
                # store the found dll
                if dll_path:
                    libs[dll.lower()] = dll_path
                # note that the dll wasn't found, so that we don't try to look for it again
                else:
                    libs[dll.lower()] = None
            # recursively process newly found dlls
            async with asyncio.TaskGroup() as group:
                for dll, dll_path, dll_deps in results:
                    if dll_path:
                        await _dep_tree(dll_path, dll_deps)

        await _dep_tree(pe, get_pe_deps(data))
        return (libs, deps)
    
    async def dep_list(path, lookup_dirs):
        result = list()
        dlls, _deps = await pe_tree(path, lookup_dirs)
        for dll, dll_path in sorted(dlls.items(), key=lambda e: e[0].casefold()):
            result.append((dll, dll_path))
        return result
else:
    import errno
        
    def normpath(path: str) -> str:
        return os.path.normpath(path).replace('//', '/')


    def readlink(path: str, root: str, prefixed: bool = False) -> str:
        root = root.rstrip('/')
        if prefixed:
            path = path[len(root):]

        while os.path.islink(root + path):
            path = os.path.join(os.path.dirname(path), os.readlink(root + path))

        return normpath((root + path) if prefixed else path)


    def dedupe(items: t.List[str]) -> t.List[str]:
        seen = {}  # type: Dict[str, str]
        return [seen.setdefault(x, x) for x in items if x not in seen]


    def parse_ld_paths(str_ldpaths, root='', path=None) -> t.List[str]:
        ldpaths = []  # type: List[str]
        for ldpath in str_ldpaths.split(':'):
            if ldpath == '':
                # The ldso treats "" paths as $PWD.
                ldpath = os.getcwd()
            elif '$ORIGIN' in ldpath:
                ldpath = ldpath.replace('$ORIGIN',
                                        os.path.dirname(os.path.abspath(path)))
            else:
                ldpath = root + ldpath
            ldpaths.append(normpath(ldpath))
        return [p for p in dedupe(ldpaths) if os.path.isdir(p)]


    @functools.lru_cache()
    def parse_ld_so_conf(ldso_conf: str,
                        root: str = '/',
                        _first: bool = True) -> t.List[str]:
        paths = []  # type: List[str]

        dbg_pfx = '' if _first else '  '
        try:
            with open(ldso_conf) as f:
                for line in f.readlines():
                    line = line.split('#', 1)[0].strip()
                    if not line:
                        continue
                    if line.startswith('include '):
                        line = line[8:]
                        if line[0] == '/':
                            line = root + line.lstrip('/')
                        else:
                            line = os.path.dirname(ldso_conf) + '/' + line
                        log.debug('%s  glob: %s', dbg_pfx, line)
                        for path in glob.glob(line):
                            paths += parse_ld_so_conf(path,
                                                    root=root,
                                                    _first=False)
                    else:
                        paths += [normpath(root + line)]
        except IOError as e:
            if e.errno != errno.ENOENT:
                log.warning(e)

        if _first:
            # XXX: Load paths from ldso itself.
            # Remove duplicate entries to speed things up.
            paths = [p for p in dedupe(paths) if os.path.isdir(p)]

        return paths


    @functools.lru_cache()
    def load_ld_paths(root: str = '/', prefix: str = '') -> t.Dict[str, t.List[str]]:
        ldpaths: t.Dict[str, t.List[str]] = {'conf': [], 'env': [], 'interp': []}

        # Load up $LD_LIBRARY_PATH.
        env_ldpath = os.environ.get('LD_LIBRARY_PATH')
        if env_ldpath is not None:
            if root != '/':
                log.warning('ignoring LD_LIBRARY_PATH due to ROOT usage')
            else:
                # XXX: If this contains $ORIGIN, we probably have to parse this
                # on a per-ELF basis so it can get turned into the right thing.
                ldpaths['env'] = parse_ld_paths(env_ldpath, path='')

        # Load up /etc/ld.so.conf.
        ldpaths['conf'] = parse_ld_so_conf(root + prefix + '/etc/ld.so.conf',
                                        root=root)
        # the trusted directories are not necessarily in ld.so.conf
        ldpaths['conf'].extend(['/lib', '/lib64/', '/usr/lib', '/usr/lib64'])
        log.debug('linker ldpaths: %s', ldpaths)
        return ldpaths


    def compatible_elfs(elf1, elf2):
        osabis = frozenset([e.header['e_ident']['EI_OSABI'] for e in (elf1, elf2)])
        compat_sets = (frozenset('ELFOSABI_%s' % x
                                for x in ('NONE',
                                        'SYSV',
                                        'GNU',
                                        'LINUX', )), )
        return ((len(osabis) == 1 or
                any(osabis.issubset(x)
                    for x in compat_sets)) and elf1.elfclass == elf2.elfclass and
                elf1.little_endian == elf2.little_endian and
                elf1.header['e_machine'] == elf2.header['e_machine'])


    def find_lib(elf, lib, ldpaths, root='/'):
        for ldpath in ldpaths:
            path = os.path.join(ldpath, lib)
            target = readlink(path, root, prefixed=True)

            if os.path.exists(target):
                with open(target, 'rb') as f:
                    libelf = elffile.ELFFile(f)
                    if compatible_elfs(elf, libelf):
                        return (target, path)

        return (None, None)


    async def lddtree(path: str,
                root: str = '/',
                prefix: str = '',
                lookup_paths: t.Optional[t.Dict[str, t.List[str]]] = None,
                display: t.Optional[str] = None,
                lib_cache: t.Dict = {},
                _first: bool = True,
                _all_libs: t.Dict[str, t.Any] = {}) -> t.Dict[str, t.Any]:
        if _first:
            _all_libs = {}
            lookup_paths = load_ld_paths().copy()
        else:
            assert lookup_paths is not None

        ret = {
            'interp': None,
            'path': path if display is None else display,
            'realpath': path,
            'needed': [],
            'rpath': [],
            'runpath': [],
            'libs': _all_libs,
        }  # type: Dict[str, Any]

        log.debug('lddtree(%s)', path)

        with open(path, 'rb') as f:
            elf = elffile.ELFFile(f)

            # If this is the first ELF, extract the interpreter.
            if _first:
                for segment in elf.iter_segments():
                    if segment.header.p_type != 'PT_INTERP':
                        continue

                    interp = segment.get_interp_name()
                    log.debug('  interp           = %s', interp)
                    ret['interp'] = normpath(root + interp)
                    ret['libs'][os.path.basename(interp)] = {
                        'path': ret['interp'],
                        'realpath': readlink(ret['interp'],
                                            root,
                                            prefixed=True),
                        'needed': [],
                    }
                    # XXX: Should read it and scan for /lib paths.
                    lookup_paths['interp'] = [
                        normpath(root + os.path.dirname(interp)),
                        normpath(root + prefix + '/usr' + os.path.dirname(
                            interp).lstrip(prefix)),
                    ]
                    log.debug('  ldpaths[interp]  = %s', lookup_paths['interp'])
                    break

            # Parse the ELF's dynamic tags.
            libs = []  # type: List[str]
            rpaths = []  # type: List[str]
            runpaths = []  # type: List[str]
            for segment in elf.iter_segments():
                if segment.header.p_type != 'PT_DYNAMIC':
                    continue

                for t in segment.iter_tags():
                    if t.entry.d_tag == 'DT_RPATH':
                        rpaths = parse_ld_paths(
                            t.rpath,
                            root=root,
                            path=path)
                    elif t.entry.d_tag == 'DT_RUNPATH':
                        runpaths = parse_ld_paths(
                            t.runpath,
                            root=root,
                            path=path)
                    elif t.entry.d_tag == 'DT_NEEDED':
                        libs.append(t.needed)
                if runpaths:
                    # If both RPATH and RUNPATH are set, only the latter is used.
                    rpaths = []

                # XXX: We assume there is only one PT_DYNAMIC.  This is
                # probably fine since the runtime ldso does the same.
                break
            if _first:
                # Propagate the rpaths used by the main ELF since those will be
                # used at runtime to locate things.
                lookup_paths['rpath'] = rpaths
                lookup_paths['runpath'] = runpaths
                log.debug('  ldpaths[rpath]   = %s', rpaths)
                log.debug('  ldpaths[runpath] = %s', runpaths)
            ret['rpath'] = rpaths
            ret['runpath'] = runpaths
            ret['needed'] = libs

            # Search for the libs this ELF uses.
            all_ldpaths = None  # type: Optional[List[str]]
            for lib in libs:
                if lib in _all_libs:
                    continue
                cached = lib_cache.get(lib, None)
                if cached is not None:
                    _all_libs[lib] = cached
                    continue

                if all_ldpaths is None:
                    all_ldpaths = (lookup_paths['rpath'] + rpaths + runpaths +
                                lookup_paths['env'] + lookup_paths['runpath'] +
                                lookup_paths['conf'] + lookup_paths['interp'])
                realpath, fullpath = find_lib(elf, lib, all_ldpaths, root)
                _all_libs[lib] = {
                    'realpath': realpath,
                    'path': fullpath,
                    'needed': [],
                }
                if fullpath:
                    lret = await lddtree(realpath,
                                root,
                                prefix,
                                lookup_paths,
                                display=fullpath,
                                lib_cache=lib_cache,
                                _first=False,
                                _all_libs=_all_libs)
                    _all_libs[lib]['needed'] = lret['needed']

            del elf

        return ret

    async def dep_list(path, lookup_dirs):
        tree = await lddtree(path, lookup_paths=lookup_dirs)
        result = list()
        for lib, data in tree['libs'].items():
            result.append((lib, data['realpath']))
        return result

async def get_runtime_dependencies(t : Executable|Library):
    lookup_dirs = t.env['PATH'].split(os.pathsep)
    return await dep_list(t.output, lookup_dirs)



from copy import deepcopy
import jinja2
from dan.core import aiofiles, asyncio
from dan.core.pathlib import Path
import re
import importlib.util

from dan.core.cache import Cache
from dan.core.find import find_file, find_files, library_paths_lookup
from dan.core.pm import re_match
from dan.core.requirements import RequiredPackage, parse_requirement
from dan.core.runners import cmdline2list
from dan.core.settings import InstallMode, InstallSettings
from dan.core.utils import unique
from dan.core.version import Version, VersionSpec
from dan.cxx.targets import CXXTarget, Library
from dan.cxx.toolchain import LibraryList

import typing as t
import os


class MissingPackage(RuntimeError):
    def __init__(self, name) -> None:
        super().__init__(f'package {name} not found')

_pkg_config_paths = None
def _get_pkg_config_paths():
    global _pkg_config_paths
    if _pkg_config_paths is None:
        paths = os.getenv('PKG_CONFIG_PATH', None)
        if paths is None:
            _pkg_config_paths = list()
        else:
            _pkg_config_paths = [Path(p) for p in paths.split(os.pathsep)]
    return _pkg_config_paths

def find_pkg_config(name, paths: str|Path|list[str|Path] = list()) -> Path:
    if isinstance(paths, (str, Path)):
        paths = [paths]
    return find_file(fr'(lib)?{re.escape(name)}\.pc$', [*paths, *_get_pkg_config_paths(), *library_paths_lookup], re.IGNORECASE)


def find_pkg_configs(name, paths: str|Path|list[str|Path] = list()) -> t.Generator[Path, None, None]:
    if isinstance(paths, (str, Path)):
        paths = [paths]
    yield from find_files(fr'(lib)?{re.escape(name)}\.pc$', [*paths, *_get_pkg_config_paths(), *library_paths_lookup], re.IGNORECASE)


def has_package(name,  paths=list()):
    return find_pkg_config(name,  paths) is not None

def parse_package_requires(reqs):
    result = []
    reqs = [r for r in re.split(r'[\s,]', reqs) if len(r.strip()) > 0]
    tmp = []
    it = iter(reqs)
    ii = next(it, None)
    while ii is not None:
        if any(op in ii for op in ('=', '>', '<')):
            tmp.append(ii)
            ii = next(it)
            tmp.append(ii)
            result.append(parse_requirement(' '.join(tmp)))
            tmp = []
        else:
            if len(tmp):
                result.append(parse_requirement(' '.join(tmp)))
                tmp = [ii]
            else:
                tmp.append(ii)
        ii = next(it, None)
    if len(tmp):
        result.append(parse_requirement(' '.join(tmp)))
    return result

class Data:
    def __init__(self, path) -> None:
        self._path = Path(path)
        self._items = dict()
        with open(self._path) as f:
            lines = [l for l in [l.strip().removesuffix('\n')
                                 for l in f.readlines()] if len(l)]
            for line in lines:
                m = Data.__split_expr.match(line)
                if m:
                    k = m.group(1).lower()
                    v = m.group(2)
                    self._items[k] = v
        self._requires = None
        self._version = None

    __split_expr = re.compile(r'(.+?)[:=](.+)')

    def get(self, name: str, default=None):
        if not name in self._items:
            return default
        value = self._items[name]
        if type(value) == str:
            while True:
                m = re.search(r'\${(\w+)}', value)
                if m:
                    var = m.group(1)
                    value = value.replace(
                        f'${{{var}}}', self.get(var))
                else:
                    break
        return value
    
    def __getstate__(self):
        return {
            'path': self._path,
            'items': self._items,
            'requires': self._requires,
            'version': self._version,
        }
    
    def __setstate__(self, data):
        self._path = data['path']
        self._items = data['items']
        self._requires = data['requires']
        self._version = data['version']

    @property
    def path(self) -> Path:
        return self._path
    
    @property
    def requires(self) -> list[RequiredPackage]:
        if self._requires is None:
            self._requires = list()
            reqs = self.get('requires')
            if reqs is not None:
                self._requires = parse_package_requires(reqs)
        return self._requires
    
    @property
    def version(self) -> Version:
        if self._version is None:
            v = self.get('version')
            if v:
                self._version = Version(v)
        return self._version


class Package(CXXTarget, internal=True):
    all: dict[str, 'Package'] = dict()

    default = False

    def __init__(self, name, search_paths: list[str] = list(), config_path: Path = None, dan_plugin=None, search_plugin=True, data: Data = None, **kwargs) -> None:
        if data is not None:
            self.config_path = data.path
            self.data = data
        else:
            self.config_path = config_path or find_pkg_config(name, search_paths)
            self.data: Data = None
        self.search_paths = search_paths
        if not self.config_path:
            raise MissingPackage(name)
        if not self.config_path.parent in self.search_paths:
            self.search_paths.insert(0, self.config_path.parent)
        self.pn = name
        self.all[name] = self

        super().__init__(f'{name}-pkgconfig', **kwargs)

        if dan_plugin:
            self.__dan_plugin = dan_plugin
        elif search_plugin:
            self.__dan_plugin = find_file(rf'{name}\.py', [self.config_path.parent.parent])
        else:
            self.__dan_plugin = None
        self.__load_plugin()
        self.__cflags = None
        self.__libs = None
        self.__lib_paths = None
        self.__bin_paths = None
        if not data:
            self.data = Data(self.config_path)


        self.version = self.data.version
    
    def __load_plugin(self):
        if self.__dan_plugin is not None:
            spec = importlib.util.spec_from_file_location(
                f'{self.name}_plugin', self.__dan_plugin)
            module = importlib.util.module_from_spec(spec)
            setattr(module, 'self', self)
            spec.loader.exec_module(module)


    def __getstate__(self):
        return {
            'pn': self.pn,
            'search_paths': self.search_paths,
            'dan_plugin': self.__dan_plugin,
            'config_path': self.config_path,
            'data': self.data,
            '__cflags': self.__cflags,
            '__libs': self.__libs,
            '__lib_paths': self.__lib_paths,
            '__bin_paths': self.__bin_paths,
        }
    
    def __setstate__(self, data: dict):
        from dan.core.include import context
        makefile = context.root
        self.__init__(data['pn'], data['search_paths'], data['config_path'],
                      data=data['data'],
                      dan_plugin=data['dan_plugin'],
                      search_plugin=False,
                      makefile=makefile)
        self.__cflags = data['__cflags']
        self.__libs = data['__libs']
        self.__lib_paths = data['__lib_paths']
        self.__bin_paths = data.get('__bin_paths', None)

    @property
    def modification_time(self):
        return 0.0

    @property
    def found(self):
        return True

    async def __initialize__(self):
        deps = set()
        reqs = self.data.requires
        if reqs:
            async with asyncio.TaskGroup(f'resolving {self.name}\'s requirements') as group:
                for req in reqs:
                    if req.name in self.all and req.is_compatible(self.all[req.name]):
                        dep = self.all[req.name]
                    else:
                        dep = find_package(req.name, req.version_spec, search_paths=self.search_paths, makefile=self.makefile)
                    if dep is None:
                        raise RuntimeError(f'Unresolved requirement: {req}')
                    group.create_task(dep.initialize())
                    deps.add(dep)
        self.includes.public.append(self.data.get('includedir'))
        self.dependencies.update(deps)

    @property
    def cxx_flags(self):
        if self.__cflags is None:
            cflags = self.data.get('cflags')
            if cflags is not None:
                cflags = cmdline2list(cflags)
                cflags = self.toolchain.from_unix_flags(cflags)
            else:
                cflags = list()
            for dep in self.cxx_dependencies:
                cflags.extend(dep.cxx_flags)
            self.__cflags = unique(cflags)
        return self.__cflags

    @property
    def package_dependencies(self):
        return [pkg for pkg in self.dependencies.all if isinstance(pkg, Package)]

    def __init_libs(self):
        self.__libs = LibraryList()
        self.__lib_paths = set()
        self.__bin_paths = set()
        libs = self.data.get('libs')
        if libs is not None:
            libs = cmdline2list(libs)
            libs = self.toolchain.from_unix_flags(libs)
            for l in libs:
                match re_match(l):
                    case r'-l(.+)' as m:
                        self.__libs.add(m[0])
                    case r'-L(.+)' as m:
                        self.__lib_paths.add(m[0])
                    case r'/LIBPATH:(.+)' as m:
                        self.__lib_paths.add(m[0])
                    case r'-Wl,-rpath,(.+)' as m:
                        self.__lib_paths.add(m[0])
                    case _:
                        self.__libs.add(l)

        bindir = self.data.get('bindir')
        if bindir is not None:
            self.__bin_paths.add(Path(bindir).as_posix())
        else:
            exec_prefix = self.data.get('exec_prefix')
            if exec_prefix is not None:
                bindir = Path(exec_prefix) / 'bin'
                if bindir.exists():
                    self.__bin_paths.add(bindir.as_posix())
        if self.toolchain.system.is_windows:
            # on windows dlls needs to be in PATH
            libdir = self.data.get('libdir')
            if libdir is not None:
                libdir = Path(libdir)
                for dll in find_files(r'.+.dll', [libdir]):
                    self.__bin_paths.add(dll.parent.as_posix())

        for pkg in self.package_dependencies:
            self.__lib_paths.update(pkg.lib_paths)
            self.__libs.extend(pkg.libs)
            self.__bin_paths.update(pkg.bin_paths)
        self.__lib_paths = list(sorted(self.__lib_paths))        

    @property
    def bin_paths(self) -> list[str]:
        if self.__bin_paths is None:
            self.__init_libs()
        return self.__bin_paths

    @property
    def lib_paths(self) -> list[str]:
        if self.__lib_paths is None:
            self.__init_libs()
        return self.__lib_paths
    
    @property
    def libs(self):
        if self.__libs is None:
            self.__init_libs()
        return self.__libs

    @asyncio.cached
    async def install(self, settings: InstallSettings, mode: InstallMode) -> list[Path]:
        settings = deepcopy(settings)
        settings.create_pkg_config = False
        return await super().install(settings, mode)

    @property
    def up_to_date(self):
        return True

    @property
    def modification_time(self):
        return self.config_path.modification_time

    def __getattr__(self, name):
        value = self.data.get(name) if self.data is not None else None
        if value is None:
            raise AttributeError(name) #@IgnoreException
        return value
    
    def __repr__(self):
        return f'Package[{self.name}] at {hex(id(self))}'


_pkgconfig_cache = None
def get_packages_cache() -> dict[str, Package]:
    from dan.core.include import context
    global _pkgconfig_cache
    if _pkgconfig_cache is None:
        _pkgconfig_cache = Cache.instance(context.root.build_path / 'pkgconfig.cache', cache_name='pkgconfig', binary=True)
    return _pkgconfig_cache.data


def find_package(name, spec: VersionSpec = None, search_paths: list = None, makefile = None):
    
    pkg = None

    if makefile is None:
        from dan.core.include import context
        makefile = context.current

    makefile = makefile.root

    cache = get_packages_cache()
    if name in cache:
        cached_pkg = cache[name]
        if spec and not spec.is_compatible(cached_pkg.version):
            raise RuntimeError(f'incompatible package {name} ({cached_pkg.version} {spec})')
        return cached_pkg

    search_paths = search_paths or [makefile.pkgs_path]
    for config in find_pkg_configs(name, search_paths):
        if spec is not None:
            data = Data(config)
            if spec.is_compatible(data.version):
                pkg = Package(name, data=data, makefile=makefile)
                break

        else:
            pkg = Package(name, config_path=config, search_paths=search_paths, makefile=makefile)
            break
    
    if pkg:
        cache[name] = pkg

    return pkg

__bindirs = None
def get_cached_bindirs():
    global __bindirs
    if __bindirs is None:
        __bindirs = set()
        for pkg in get_packages_cache().values():
            __bindirs.update(pkg.bin_paths)

    return __bindirs

_jinja_env: jinja2.Environment = None
def _get_jinja_env():
    global _jinja_env
    if _jinja_env is None:
        _jinja_env = jinja2.Environment(
            loader=jinja2.PackageLoader('dan.pkgconfig'))
    return _jinja_env


async def create_pkg_config(lib: Library, settings: InstallSettings) -> Path:
    dest = settings.data_destination / 'pkgconfig' / f'{lib.name}.pc'
    lib.info(f'creating pkgconfig: {dest}')

    requires = list()
    for dep in lib.dependencies.public:
        match dep:
            case RequiredPackage():
                if dep.version_spec:
                    requires.append(f'{dep.name} {dep.version_spec.op} {dep.version_spec.version}')
                else:
                    requires.append(dep.name)
            case Library():
                requires.append(dep.name)


    libs = list()
    if not lib.interface:
        libs.extend(lib.toolchain.make_libpath_options(
            [Path(f'${{libdir}}/{lib.name}')]))
        libs.extend(lib.toolchain.make_link_options(
            [Path(f'${{libdir}}/{lib.name}')]))
    for p in lib.library_paths.public:
        if p not in libs:
            libs.append(p)
    libs.extend(lib.link_libraries.public)
    libs.extend(lib.link_options.public)
    libs = lib.toolchain.to_unix_flags(libs)

    cflags = lib.compile_definitions.public
    cflags.extend(lib.compile_options.public)
    cflags.extend(lib.toolchain.make_include_options(['${includedir}']))
    cflags = lib.toolchain.to_unix_flags(cflags)

    data = _get_jinja_env()\
        .get_template('pkg.pc.jinja2')\
        .render({
            'lib': lib,
            'libs': libs,
            'cflags': cflags,
            'settings': settings,
            'prefix': Path(settings.destination).absolute(),
            'requires': requires
        })
    dest.parent.mkdir(parents=True, exist_ok=True)
    async with aiofiles.open(dest, 'w') as f:
        await f.write(data)
    return dest



import tqdm
from enum import Enum

class ProgressMode(Enum):
    NONE = 0
    BASIC = 1
    IDENTIFIED = 2

mode = ProgressMode.BASIC

def set_progress_mode(new_mode: ProgressMode):
    global mode
    mode = new_mode

class progress(tqdm.tqdm):

    __uid_count = 1

    __doc__ = tqdm.tqdm.__doc__
    def __init__(self, *args, **kwargs):
        if kwargs.pop("root", False):
            self.__id = 0
        else:
            self.__id = self.__uid_count
            progress.__uid_count += 1
        desc = kwargs.pop("desc", "")
        if mode == ProgressMode.IDENTIFIED:
            desc = f'{self.__id}-{desc}'
        super().__init__(*args, desc=desc, **kwargs)
        self.disable = mode == ProgressMode.NONE
    
    def set_description(self, desc="", refresh=True):
        if mode == ProgressMode.IDENTIFIED:
            desc = f'{self.__id}-{desc}'
        super().set_description(desc, refresh)

    def set_description_str(self, desc="", refresh=True):
        if mode == ProgressMode.IDENTIFIED:
            desc = f'{self.__id}-{desc}'
        super().set_description(desc, refresh)
    
    def close(self) -> None:
        if mode == ProgressMode.IDENTIFIED and self.total == None:
            # ensure a 100% is printed
            self.total = self.n
            self.n -= 1
            self.update()
            self.refresh()
        return super().close()

    def __call__(self, *args, **kwargs):
        super().update(*args, **kwargs)

progress.__init__.__doc__ = tqdm.tqdm.__init__.__doc__



import platform
from typing import Any
from dan.core.pm import re_match
from dan.core.runners import async_run
from dan.core.target import Target
from dan.core import aiofiles
import jinja2

_conanfile_template = """[requires]
{% for r in requirements %}
{{r.name}}/{{r.version}}
{%- endfor %}

[options]
{% for r in requirements -%}
{% for k, v in r.options.items() %}
{{r.name}}/*:{{k}}={{v}}
{%- endfor %}
{%- endfor %}

[generators]
PkgConfigDeps
"""


class ConanFile(Target, internal=True):
    def __init__(self, makefile) -> None:
        super().__init__('conanfile', makefile=makefile)
        self.__reqs = list()
        self.output = 'conanfile.txt'

    def add(self, pkg: 'Package'):
        self.__reqs.append(pkg)

    async def __build__(self):
        template = jinja2.Environment(
            loader=jinja2.BaseLoader).from_string(_conanfile_template)
        content = template.render(requirements=self.__reqs)
        async with aiofiles.open(self.output, 'w') as out:
            await out.write(content)


class Requirements(Target, internal=True):

    def __init__(self, makefile) -> None:
        super().__init__('conan', makefile=makefile)

        self.output = self.makefile.pkgs_path / 'lib' / 'pkgconfig' / 'conanrun'
        if platform.system() == 'Windows':
            self.output = self.output.with_suffix('.bat')
        else:
            self.output = self.output.with_suffix('.sh')

        self.conanfile = ConanFile(makefile=makefile)
        self.preload_dependencies.add(self.conanfile)

        self.toolchain = self.context.get('cxx_target_toolchain')

    def _get_version(self, toolchain):
        from dan.cxx.msvc_toolchain import MSVCToolchain
        match toolchain:
            case MSVCToolchain():
                return f'{toolchain.version.major}{str(toolchain.version.minor)[0]}'
            case _:
                return str(toolchain.version.major)
            
    def add(self, pkg: 'Package'):
        self.conanfile.add(pkg)

    async def __build__(self):
        dest = self.output.parent
        dest.mkdir(exist_ok=True, parents=True)
        await async_run([
            'conan', 'install', '.',
            f'--output-folder={dest}',
            '-s', f'build_type={self.toolchain.build_type.name.title()}',
            '-s', f'compiler={self.toolchain.type}',
            '-s', f'compiler.version={self._get_version(self.toolchain)}',
            '-s', f'compiler.cppstd={self.toolchain.cpp_std}',
            '--build=missing'],
            logger=self._logger, cwd=self.build_path)

class Package(Target, internal=True):
    def __get_requirements(self) -> Requirements:
        if not hasattr(self.makefile, '__conan_requirements__'):
            setattr(self.makefile, '__conan_requirements__', Requirements(makefile=self.makefile))
        return getattr(self.makefile, '__conan_requirements__')

    def __init__(self):
        super().__init__()
        # assert self.parent is not None
        reqs = self.__get_requirements()
        reqs.add(self)
        self.output = reqs.output.parent / f'{self.name}.pc'
        self.dependencies.add(reqs)



from dataclasses import dataclass, field
from enum import Enum
import re

from dan.core.pathlib import Path
import dan.core.typing as t


class InstallMode(Enum):
    user = 0
    dev = 1
    portable = 2


class BuildType(Enum):
    debug = 0
    release = 1
    release_min_size = 2
    release_debug_infos = 3

    @property
    def is_debug_mode(self):
        """Return true if the build type should produce debug symbols (ie.: debug and release_debug_infos)"""
        return self in (BuildType.debug, BuildType.release_debug_infos)
    
class DefaultLibraryType(Enum):
    static = 0
    shared = 1

@dataclass(eq=True, unsafe_hash=True)
class InstallSettings:
    destination: str = '/usr/local'
    runtime_prefix: str = 'bin'
    libraries_prefix: str = 'lib'
    includes_prefix: str = 'include'
    data_prefix:str = 'share'
    create_pkg_config: bool = True

    @property
    def runtime_destination(self):
        return Path(self.destination).absolute() / self.runtime_prefix

    @property
    def libraries_destination(self):
        return Path(self.destination).absolute() / self.libraries_prefix

    @property
    def data_destination(self):
        return Path(self.destination).absolute() / self.data_prefix

    @property
    def includes_destination(self):
        return Path(self.destination).absolute() / self.includes_prefix

@dataclass
class ToolchainSettings:
    cxx_flags: list[str] = field(default_factory=lambda: list())
    default_library_type: DefaultLibraryType = DefaultLibraryType.static

@dataclass
class Settings:
    build_type: BuildType = BuildType.debug
    install: InstallSettings = field(default_factory=lambda: InstallSettings())
    target: ToolchainSettings = field(default_factory=lambda: ToolchainSettings())


def safe_load(name: str, value,  t: type):
    if t is not None and not isinstance(value, t):
        err = f'value {name} should be of type {t}'
        if type(value) == str:
            if issubclass(t, Enum):
                names = [n.lower()
                         for n in t._member_names_]
                value = value.lower()
                if value in names:
                    value = t(names.index(value))
                else:
                    err = f'option {name} should be one of {names}'
            else:
                import json
                value = json.loads(value)
            if not isinstance(value, t):
                raise RuntimeError(err)
        else:
            raise RuntimeError(err)
    return value

def _parse_str_value(name, value: str, orig: type, tp: type = None):
    if issubclass(orig, Enum):
        names = [n.lower()
                    for n in orig._member_names_]
        value = value.lower()
        if value in names:
            return orig(names.index(value))
        else:
            raise RuntimeError(f'{name} should be one of {names}')
    elif issubclass(orig, (set, list)):
        assert tp is not None
        result = list()
        for sub in value.split(';'):
            result.append(_parse_str_value(name, sub, tp))
        return orig(result)
    elif orig == bool:
        return value.lower() in ('true', 'yes', 'on', '1')
    else:
        if tp is not None:
            raise TypeError(f'unhandled type {orig}[{tp}]')
        return orig(value)

def _apply_inputs(inputs: list[str], get_item: t.Callable[[str], tuple[t.Any, t.Any, t.Any]], logger = None, input_type_name='setting'):
    for input in inputs:
        m = re.match(r'(.+?)([+-])?="?(.+)"?', input)
        if m:
            name = m[1]
            op = m[2]
            value = m[3]
            _item = get_item(name)
            if _item is None and logger is not None:
                logger.warning('unknown %s: %s (skipped)', input_type_name, name)
                continue
            input, out_value, orig = _item
            sname = name.split('.')[-1]
            if orig is None:
                orig = type(input)
            if hasattr(orig, '__annotations__') and sname in orig.__annotations__:
                tp = orig.__annotations__[sname]
                if t.is_optional(tp):
                    orig = t.get_args(tp)[0]
                    tp = None
                else:
                    orig = t.get_origin(tp)
                    if orig is None:
                        orig = tp
                        tp = None
                    else:
                        args = t.get_args(tp)
                        if args:
                            tp = args[0]
            else:
                tp = None
            in_value = _parse_str_value(name, value, orig, tp)
            match (out_value, op, in_value):
                case (list()|set(), '-', list()|set()) if len(in_value) == 1 and list(in_value)[0] == '*':
                    out_value.clear()
                case (set(), '+', set()):
                    out_value.update(in_value)
                case (set(), '-', set()):
                    out_value = out_value - in_value
                case (list(), '+', set()|list()):
                    out_value.extend(in_value)
                case (list(), '-', set()|list()):
                    for v in in_value:
                        out_value.remove(v)
                case (_, '+' | '-', _):
                    raise TypeError(f'unhandled "{op}=" operator on type {type(out_value)} ({name})')
                case _:
                    out_value = in_value
            if isinstance(input, dict):
                input[sname] = out_value
            else:
                setattr(input, sname, out_value)
            if logger is not None:
                logger.info('%s: %s = %s', input_type_name, name, out_value)
        else:
            raise RuntimeError(f'cannot process given input: {input}')


def apply_settings(base, *settings, logger=None):
    def get_setting(name):
        parts = name.split('.')
        setting = base
        for part in parts[:-1]:
            if not hasattr(setting, part):
                raise RuntimeError(f'no such setting: {name}')
            setting = getattr(setting, part)
        if not hasattr(setting, parts[-1]):
            raise RuntimeError(f'no such setting: {name}')
        value = getattr(setting, parts[-1])
        return setting, value, type(setting)
    _apply_inputs(settings, get_setting, logger=logger)



import asyncio
from aiofiles import *
from aiofiles import os

import os as sync_os
import stat
import re
import sys
import errno
import contextlib
import time

from dan.core.pathlib import Path

async def _rmtree(path: Path):
    for root, dirs, files in sync_os.walk(path, topdown=False):
        clean_files = [os.remove(
            sync_os.path.join(root, name)) for name in files]
        clean_dirs = [os.rmdir(
            sync_os.path.join(root, name)) for name in dirs]
        await asyncio.gather(*clean_files)
        await asyncio.gather(*clean_dirs)
    await os.rmdir(path)

async def _remove_force(path: Path):
    sync_os.chmod(path, stat.S_IWRITE)
    await os.remove(path)

async def _rmtree_force(path: Path):
    for root, dirs, files in sync_os.walk(path, topdown=False):
        clean_files = [_remove_force(
            Path(root) / name) for name in files]
        clean_dirs = [os.rmdir(
            sync_os.path.join(root, name)) for name in dirs]
        await asyncio.gather(*clean_files)
        await asyncio.gather(*clean_dirs)
    await os.rmdir(path)

async def rmtree(path: Path, force=False):
    if force:
        await _rmtree_force(path)
    else:
        await _rmtree(path)

async def copy(src : Path, dest : Path, chunk_size=2048):
    if dest.is_dir():
        dest = dest / src.name
    dest.parent.mkdir(exist_ok=True, parents=True)
    async with open(src, 'rb') as s,\
               open(dest, 'wb') as d:
        while True:
            chunk = await s.read(chunk_size)
            if len(chunk) == 0:
                break
            await d.write(chunk)
    dest.chmod(src.stat().st_mode)


async def sub(filepath, pattern, repl, **kwargs):
    async with open(filepath) as f:
        content = await f.read()
    content = re.sub(pattern, repl, content, **kwargs)
    async with open(filepath, 'w') as f:
        await f.write(content)



class FileLock:
    def __init__(self, path: str|Path, timeout=None, poll_interval=0.1) -> None:
        self._path = Path(path)
        self._mode: int = 0o644
        self._fh = None
        self._timeout = timeout
        self._poll_interval = poll_interval

    def __del__(self):
        if self.has_lock:
            self.release()

    @property
    def locked(self):
        return self._path.exists()

    @property
    def has_lock(self):
        return self._fh is not None

    def try_acquire(self):
        flags = (
            sync_os.O_WRONLY  # open for writing only
            | sync_os.O_CREAT
            | sync_os.O_EXCL  # together with above raise EEXIST if the file specified by filename exists
            | sync_os.O_TRUNC  # truncate the file to zero byte
        )
        try:
            self._fh = sync_os.open(self._path, flags, self._mode)
            return True
        except OSError as exception:  # re-raise unless expected exception
            if not (
                exception.errno == errno.EEXIST  # lock already exist
                or (exception.errno == errno.EACCES and sys.platform == "win32")  # has no access to this lock
            ):  # pragma: win32 no cover
                raise
            return False

    async def acquire(self, timeout=None):
        if timeout is None:
            timeout = self._timeout
        t0 = t1 = time.perf_counter()
        while timeout is None or timeout < t1 - t0:
            if self.try_acquire():
                return True
            await asyncio.sleep(self._poll_interval)
            t1 = time.perf_counter()
        return False

    def release(self):
        assert self._fh is not None
        sync_os.close(self._fh)
        self._fh = None
        with contextlib.suppress(OSError):  # the file is already deleted and that's what we want
            self._path.unlink()

    async def __aenter__(self):
        await self.acquire()
    
    async def __aexit__(self, *exc):
        self.release()



import os
import shutil
import tempfile
from pathlib import Path
from dan.core.pathlib import Path
from dan.core.target import Target
from dan.core import aiofiles, asyncio
from dan.core.runners import async_run
from dan.src.base import SourcesProvider
import tarfile
import zipfile

from dan.utils.net import fetch_file

class TarSources(SourcesProvider, internal=True):

    url: str
    archive_name: str = None
    extract_filter = None
    patches = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.output = str(self.name)

    def __extract__(self, archive_path: Path, dest: Path):
        if archive_path.suffix == '.zip':
            with zipfile.ZipFile(archive_path) as f:
                root = os.path.commonprefix(f.namelist())
                f.extractall(dest)
        else:
            mode = 'r:*'
            if len(archive_path.suffixes) and archive_path.suffixes[-1] == '.xz':
                mode = 'r:xz'
            with tarfile.open(archive_path, mode) as f:
                root = None
                def _filter(member: tarfile.TarInfo):
                    nonlocal root
                    if root is None:
                        root = member.name
                    else:
                        root = os.path.commonprefix([root, member.name])
                    if self.extract_filter is None:
                        return member
                    else:
                        return self.extract_filter(member, dest / member.name)
                members = []
                for m in f:
                    m = _filter(m)
                    if m:
                        members.append(m)
                self.debug('extracting %d members', len(members))
                f.extractall(dest, members)
        return root

    async def __build__(self):
        archive_name = self.archive_name or self.url.split("/")[-1]
        archive_path = self.build_path / archive_name
        if archive_path.exists():
            self.debug('%s already available (download skipped)', archive_path)
        else:
            self.info(f'downloading {self.url}')
            await fetch_file(self.url, self.build_path / archive_name, self.name)
        with tempfile.TemporaryDirectory(prefix=f'{self.name}-') as tmp_dest:
            extract_dest = Path(tmp_dest) / 'a'
            self.info(f'extracting {archive_name}')
            root = await asyncio.get_event_loop().run_in_executor(None, self.__extract__, archive_path, extract_dest)
            await asyncio.get_event_loop().run_in_executor(None, shutil.move, extract_dest / root, self.output)
            
            if self.patches is not None:
                for patch in self.patches:
                    self.info('applying %s', patch)
                    await async_run(['bash', '-c', 'cd', self.output , '&&', 'patch', '-p0', '<', self.source_path / patch], logger=self, cwd=self.output)

            await aiofiles.os.remove(archive_path)




import functools
import typing as t
from dan.core import asyncio
from dan.core.pm import re_match
from dan.core.settings import InstallMode, InstallSettings

from dan.core.version import VersionSpec
from dan.logging import Logging

def parse_package(name: str) -> tuple[str, str, str]:
    """Parse package name
    
    :returns: package, library, repository"""
    match re_match(name):
        # full specification <pkg>:<lib>@<repo>
        case r'(.+?):(.+?)@(.+)' as m:
            package = m[1]
            library = m[2]
            repository = m[3]
        # repo specification <lib>@<repo>
        case r'(.+?)@(.+)' as m:
            package = None
            library = m[1]
            repository = m[2]
        # package specification <pkg>:<lib>
        case r'(.+?):(.+)' as m:
            package = m[1]
            library = m[2]
            repository = None
        # no specification, automatic resolution in default repository
        case _:
            package = None
            library = name
            repository = None
    
    return package, library, repository


class RequiredPackage(Logging):
    
    @t.overload
    def __init__(self, specification_str: str): ...
    
    @t.overload
    def __init__(self, name: str, version_spec: VersionSpec): ...

    def __init__(self, *args):
        if len(args) == 2:
            name = args[0]
            self.version_spec = args[1]
        else:
            name, self.version_spec = VersionSpec.parse(args[0])
        self.target : 'Target' = None
        self.pn = name
        self.package, self.name, self.repository = parse_package(name)
        # fullfill Logging requirements
        setattr(self, 'fullname', str(self))
        setattr(self, '_logger', None)
        self.__skipped = list()

    def __getstate__(self) -> object:
        return {
            'pn': self.pn,
            'version_spec': self.version_spec,
        }
    
    def __setstate__(self, data):
        self.__init__(data['pn'], data['version_spec'])


    def is_compatible(self, t: 'Target'):
        if self.version_spec is not None:
            version_ok = self.version_spec.is_compatible(t.version)
        else:
            version_ok = True
        return version_ok
    
    @property
    def found(self):
        return self.target is not None
    
    @property
    def modification_time(self):
        return self.target.modification_time if self.target else 0.0

    def __skipped_method_call(self, name, *args, **kwargs):
        if name not in self.__skipped:
            self.debug('call to %s skipped (unresolved)', name)
            self.__skipped.append(name)

    def __getattr__(self, name):
        if not self.found:
            return functools.partial(self.__skipped_method_call, name)
        else:
            return getattr(self.target, name)
    
    def __str__(self) -> str:
        if self.version_spec:
            return f'RequiredPackage[{self.name} {self.version_spec}]'
        return f'RequiredPackage[{self.name}]'
    
    def __repr__(self) -> str:
        return f'{self} at {hex(id(self))}'
    

def parse_requirement(req: str) -> RequiredPackage:
    req = req.strip()
    name, spec = VersionSpec.parse(req)
    if spec:
        return RequiredPackage(name, spec)
    else:
        return RequiredPackage(req)

async def load_requirements(requirements: t.Iterable[RequiredPackage], makefile, logger = None, install = True):

    from dan.pkgconfig.package import find_package
    from dan.logging import _get_makefile_logger
    from dan.io import Package

    if logger is None:
        logger = _get_makefile_logger()

    deps_install_path = makefile.root.pkgs_path
    deps_settings = InstallSettings(deps_install_path)

    pkgs_search_paths = [deps_install_path]
    if makefile.requirements:
        pkgs_search_paths.append(makefile.requirements.pkgs_path)

    resolved: list[RequiredPackage] = list()
    unresolved: list[RequiredPackage] = list()

    async with asyncio.TaskGroup('requirement loading') as group:
        for req in requirements:
            if req.found:
                resolved.append(req)
                continue

            t = makefile.context.root.find(req.name)
            if t and not t.is_requirement and req.is_compatible(t):
                logger.debug('%s: already fullfilled by %s', req, t.fullname)
                req.target = t
                resolved.append(req)
                continue

            t = find_package(req.name, req.version_spec, search_paths=pkgs_search_paths, makefile=makefile)
            if t is not None and req.is_compatible(t):
                logger.debug('%s: using package %s', req, t.fullname)
                req.target = t
                resolved.append(req)
                continue
            
            if makefile.requirements:
                # install requirement from dan-requires.py
                t = makefile.requirements.find(req.name)
                if not t:
                    raise RuntimeError(f'Unresolved requirement {req}, it should have been defined in {makefile.requirements.__file__}')
                logger.debug('%s using requirements\' target %s', req, t.fullname)
            else:
                with makefile.context:
                    t, is_new = await Package.instance(req.name, req.version_spec, package=req.package, repository=req.repository, makefile=makefile.root)
                if is_new:
                    logger.debug('%s: adding package %s', req, t.fullname)
                elif install:
                    logger.debug('%s: package already beeing installed at version %s', req, t.version)

            if install:
                group.create_task(t.install(deps_settings, InstallMode.dev))
            unresolved.append(req)

    if install:
        for req in unresolved:
            pkg = find_package(req.name, req.version_spec, search_paths=pkgs_search_paths, makefile=makefile)
            if pkg is None:
                raise RuntimeError(f'Unresolved requirement {req}')
            req.target = pkg
            resolved.append(req)

    return [req.target for req in resolved]
